# Copyright 2008-2015 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.delegates.
#
# lazr.delegates is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.delegates is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.delegates.  If not, see <http://www.gnu.org/licenses/>.

"""Decorator helpers that simplify class composition."""

from __future__ import absolute_import, print_function, unicode_literals


__metaclass__ = type
__all__ = [
    'delegates',
    ]


import sys
from types import ClassType

from zope.interface.advice import addClassAdvisor
from zope.interface import classImplements
from zope.interface.interfaces import IInterface

from lazr.delegates._passthrough import Passthrough


def delegates(interface_spec, context='context'):
    """Make an adapter into a decorator.

    Use like:

        class RosettaProject:
            implements(IRosettaProject)
            delegates(IProject)

            def __init__(self, context):
                self.context = context

            def methodFromRosettaProject(self):
                return self.context.methodFromIProject()

    If you want to use a different name than "context" then you can explicitly
    say so:

        class RosettaProject:
            implements(IRosettaProject)
            delegates(IProject, context='project')

            def __init__(self, project):
                self.project = project

            def methodFromRosettaProject(self):
                return self.project.methodFromIProject()

    The adapter class will implement the interface it is decorating.

    The minimal decorator looks like this:

    class RosettaProject:
        delegates(IProject)

        def __init__(self, context):
            self.context = context

    """
    # pylint: disable-msg=W0212
    frame = sys._getframe(1)
    locals = frame.f_locals

    # Try to make sure we were called from a class def
    if (locals is frame.f_globals) or ('__module__' not in locals):
        raise TypeError(
            "delegates() can be used only from a class definition.")

    locals['__delegates_advice_data__'] = interface_spec, context
    addClassAdvisor(_delegates_advice, depth=2)


def _delegates_advice(cls):
    """Add a Passthrough class for each missing interface attribute.

    This function connects the decorator class to the delegate class.
    Only new-style classes are supported.
    """
    interface_spec, contextvar = cls.__dict__['__delegates_advice_data__']
    del cls.__delegates_advice_data__
    if isinstance(cls, ClassType):
        raise TypeError(
            'Cannot use delegates() on a classic class: %s.' % cls)
    if IInterface.providedBy(interface_spec):
        interfaces = [interface_spec]
    else:
        interfaces = interface_spec
    not_found = object()
    for interface in interfaces:
        classImplements(cls, interface)
        for name in interface:
            if getattr(cls, name, not_found) is not_found:
                setattr(cls, name, Passthrough(name, contextvar))
    return cls



# Copyright 2009-2015 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.delegates
#
# lazr.delegates is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.delegates is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.delegates.  If not, see <http://www.gnu.org/licenses/>.

"""Doctest fixtures for running under nose."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'globs',
    ]


def globs(globs):
    """Set up globals for doctests."""
    # Enable future statements to make Python 2 act more like Python 3.
    globs['absolute_import'] = absolute_import
    globs['print_function'] = print_function
    globs['unicode_literals'] = unicode_literals
    # Provide a convenient way to clean things up at the end of the test.
    return globs



# Copyright 2014-2015 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.delegates.
#
# lazr.delegates is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.delegates is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.delegates.  If not, see <http://www.gnu.org/licenses/>.

"""Class decorator definition for Python 3.

This syntax is not legal in Python 2.
"""

from zope.interface import classImplements

from lazr.delegates._passthrough import Passthrough


def delegate_to(*interfaces, context='context'):
    if len(interfaces) == 0:
        raise TypeError('At least one interface is required')
    def _decorator(cls):
        missing = object()
        for interface in interfaces:
            classImplements(cls, interface)
            for name in interface:
                if getattr(cls, name, missing) is missing:
                    setattr(cls, name, Passthrough(name, context))
        return cls
    return _decorator



# Copyright 2013-2015 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.delegates.
#
# lazr.delegates is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.delegates is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.delegates.  If not, see <http://www.gnu.org/licenses/>.

"""Test the legacy API, which only works in Python 2.

All of these tests are copied almost verbatim from the old README.rst.
"""


from __future__ import absolute_import, print_function, unicode_literals


# Don't enable the following or we can't test classic class failures.
#__metaclass__ = type
__all__ = [
    'TestLegacyAPI',
    ]


import sys
import unittest

from zope.interface import Attribute, Interface, implements, providedBy

from lazr.delegates import delegate_to
if sys.version_info[0] == 2:
    from lazr.delegates import delegates


class IFoo0(Interface):
    spoo = Attribute('attribute in IFoo0')

class IFoo(IFoo0):
    def bar():
        'some method'
    baz = Attribute('some attribute')

class BaseFoo0:
    spoo = 'some spoo'

class BaseFoo(BaseFoo0):
    def bar(self):
        return 'bar'
    baz = 'hi baz!'

class IOther(Interface):
    another = Attribute('another attribute')

class BaseOtherFoo(BaseFoo):
    another = 'yes, another'


# Python 2.6 doesn't have skips.
def skip_python3(cls):
    if sys.version_info[0] > 2:
        return None
    return cls


@skip_python3
class TestLegacyAPI(unittest.TestCase):
    def test_basic_usage(self):
        class SomeClass(object):
            delegates(IFoo)
            def __init__(self, context):
                self.context = context

        f = BaseFoo()
        s = SomeClass(f)
        self.assertEqual(s.bar(), 'bar')
        self.assertEqual(s.baz, 'hi baz!')
        self.assertEqual(s.spoo, 'some spoo')
        self.assertTrue(IFoo.providedBy(s))

    def test_keyword_context(self):
        class SomeOtherClass(object):
            delegates(IFoo, context='myfoo')
            def __init__(self, foo):
                self.myfoo = foo
            spoo = 'spoo from SomeOtherClass'

        f = BaseFoo()
        s = SomeOtherClass(f)
        self.assertEqual(s.bar(), 'bar')
        self.assertEqual(s.baz, 'hi baz!')
        self.assertEqual(s.spoo, 'spoo from SomeOtherClass')

        s.baz = 'fish'
        self.assertEqual(s.baz, 'fish')
        self.assertEqual(f.baz, 'fish')

    def test_classic_is_error(self):
        try:
            class SomeClassicClass:
                delegates(IFoo)
        except TypeError:
            pass
        else:
            self.fail('TypeError expected')

    def test_use_outside_class_is_error(self):
        self.assertRaises(TypeError, delegates, IFoo)

    def test_multiple_interfaces(self):
        class SomeOtherClass(object):
            delegates([IFoo, IOther])

        s = SomeOtherClass()
        s.context = BaseOtherFoo()
        self.assertEqual(s.another, 'yes, another')
        self.assertEqual(s.baz, 'hi baz!')
        self.assertEqual(s.spoo, 'some spoo')
        self.assertTrue(IFoo.providedBy(s))
        self.assertTrue(IOther.providedBy(s))

    def test_decorate_existing_object(self):
        class MoreFoo(BaseFoo, BaseOtherFoo):
            implements([IFoo, IOther])

        foo = MoreFoo()

        class WithExtraTeapot(object):
            delegates(providedBy(foo))
            teapot = 'i am a teapot'

        foo_with_teapot = WithExtraTeapot()
        foo_with_teapot.context = foo

        self.assertEqual(foo_with_teapot.baz, 'hi baz!')
        self.assertEqual(foo_with_teapot.another, 'yes, another')
        self.assertEqual(foo_with_teapot.teapot, 'i am a teapot')
        self.assertTrue(IFoo.providedBy(foo_with_teapot))
        self.assertTrue(IOther.providedBy(foo_with_teapot))


@skip_python3
class TestNewAPI(unittest.TestCase):
    """Test corner cases in Python 2.

    Most of the new API is tested in the doctest.  The implementation of the
    new API is different between Python 2 and Python 3, so test these corner
    cases.
    """
    def test_type_error(self):
        # Too many arguments to @delegate_to() raises a TypeError.
        try:
            @delegate_to(IFoo0, context='myfoo', other='bogus')
            class SomeClass(object):
                pass
        except TypeError:
            pass
        else:
            self.fail('TypeError expected')



# Copyright 2008-2015 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.delegates.
#
# lazr.delegates is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.delegates is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.delegates.  If not, see <http://www.gnu.org/licenses/>.

"""Decorator helpers that simplify class composition."""

__all__ = [
    'delegate_to',
    ]


import pkg_resources
__version__ = pkg_resources.resource_string(
    "lazr.delegates", "version.txt").strip()

from lazr.delegates._passthrough import Passthrough

# The class decorator syntax is different in Python 2 vs. Python 3.
import sys
if sys.version_info[0] == 2:
    from lazr.delegates._python2 import delegate_to
    # The legacy API is only compatible with Python 2.
    from lazr.delegates._delegates import delegates
    __all__.append('delegates')
else:
    from lazr.delegates._python3 import delegate_to



# Copyright 2008-2015 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.delegates.
#
# lazr.delegates is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.delegates is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.delegates.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

__version__ = open("lazr/delegates/version.txt").read().strip()

setup(
    name='lazr.delegates',
    version=__version__,
    namespace_packages=['lazr'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    maintainer='LAZR Developers',
    maintainer_email='lazr-developers@lists.launchpad.net',
    description='Easily write objects that delegate behavior',
    long_description="""
The ``lazr.delegates`` package makes it easy to write objects that delegate
behavior to another object. The new object adds some property or behavior on
to the other object, while still providing the underlying interface, and
delegating behavior.
""",
    license='LGPL v3',
    install_requires=[
        'nose',
        'setuptools',
        'zope.interface',
        ],
    url='https://launchpad.net/lazr.delegates',
    download_url= 'https://launchpad.net/lazr.delegates/+download',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        ],
    # nose plugins don't really work with `python setup.py test` so use
    # `python setup.py nosetests` instead, or just `tox`.  Gosh, we really
    # should switch to nose2. :/  - BAW 2014-08-20
    #test_suite='nose.collector',
    )
